/**
 * Not the real type definitions, but just enough for this client.
 * @see https://pics.me.me/duck-typing-duck-typing-2429782.png - an accurate representation of this module.
 */
export as namespace L;

export function geoJSON(...any): any;
export function tileLayer(...any): any;
export function map(...any): any;
export function latLngBounds(coord1: Array<number>, coord2: Array<number>): any;
